@@ -4,25 +4,45 @@ use 5.008;
 use strict;
 use warnings;
 no warnings qw(uninitialized);
+use Carp;
+use File::Spec;
 
-our $VERSION = '0.010';
+our $VERSION = '0.020';
 $VERSION = eval $VERSION;
 
 sub import {
     my $package = shift;
     for my $category (@_) {
-        enable_warning_category($category);
+        if (!ref($category)) {
+            enable_warning_category($category);
+        }
     }
 }
 
 sub unimport {
     my $package = shift;
-    for my $category (@_) {
-        disable_warning_category($category);
+    for my $args (@_) {
+        if (ref($args)) {
+            $package->_check_import_argument($args);
+            $package->_thwart_module(%$args);
+        } else {
+            disable_warning_category($args);
+        }
     }
 }
 
-use Carp;
+sub _check_import_argument {
+    my ($package, $argument) = @_;
+
+    return if !ref($argument);
+    if (ref($argument) ne 'HASH') {
+        croak "Unexpected import argument $argument";
+    }
+    if (!exists $argument->{warning} || !exists $argument->{thwart_module}) {
+        croak "Argument keys must include warning and thwart_module";
+    }
+    _check_warning_category($argument->{warning}) or die;
+}
 
 =head1 NAME
 
@@ -30,16 +50,37 @@ warnings::everywhere - a way of ensuring consistent global warning settings
 
 =head1 VERSION
 
-This is version 0.007.
+This is version 0.020.
 
 =head1 SYNOPSIS
 
+ # Turn off run-time warnings
+ use strict;
+ use warnings;
  no warnings::anywhere qw(uninitialized);
+ 
  use Module::That::Spits::Out::Warnings;
  use Other::Unnecessarily::Chatty::Module;
+
  use warnings::everywhere qw(uninitialized);
  # Write your own bondage-and-discipline code that really, really
  # cares about the difference between undef and the empty string
+ 
+ # Stop "helpful" modules from turning compile-time warnings back on again
+ use strict;
+ use warnings;
+ no warnings::anywhere {
+     warning       => 'experimental::smartmatch',
+     thwart_module => [qw(Moose Moo Dancer Dancer2)],
+ };
+ use Module::That::Might::Pull::In::Moose::Or::Moo::Or::Who::Knows::What;
+ 
+ given (shift @ARGV) {
+     ...
+     default {
+         print STDERR "# I'll fix it in a moment, OK?\n";
+     }
+ }
 
 =head1 DESCRIPTION
 
@@ -73,10 +114,18 @@ they were defined first.
 Either way, this isn't my code, and it's not something I'm going to fix.
 These warnings are just spam.
 
+Similarly, if you disable e.g. experimental::smartmatch because you know that
+you're using smartmatch, and you're not going to be using a version of
+Perl that has a version of smartmatch that behaves differently, you might
+get those warnings enabled back again by a module such as Moose or Dancer
+which turns all warnings on.
+
 This is where warnings::everywhere comes in.
 
 =head2 Usage
 
+=head3 Run-time warnings
+
 At its simplest, say
 
  use warnings::everywhere qw(all);
@@ -99,6 +148,27 @@ or
 Parameters are the same as C<use warnings>: a list of categories
 as per L<perllexwarn>, where C<all> means all warnings.
 
+=head3 Compile-time warnings
+
+This won't work for some (all?) compile-time warnings that are not just
+enabled for the module in question, but are injected back into your package.
+Moose, Moo, Dancer and Dancer2 all do this at the time of writing, by saying
+C<warnings->import> in their import method, thus injecting all warnings into
+I<your> package.
+
+To stop such code from turning back on warnings that you thought you'd
+disabled, say e.g.
+
+ no warnings::anywhere {
+     warning       => 'experimental::smartmatch',
+     thwart_module => [qw(Moose)],
+ };
+
+B<Warning>: warnings::everywhere disables these warnings by what is basically
+a source filter, so use with caution. If you can find an approved way of
+preventing modules such as Moose from doing this, do that rather than
+messing about with the module's source code!
+
 =head2 Limitations
 
 warnings::everywhere works by fiddling with the contents of the global hashes
@@ -162,12 +232,12 @@ after all. So you can't say
 and expect to get a warning from the last line. That warning's been
 turned off for good.
 
-=item It won't work for compile-time warnings
+=item Its method of disabling compile-time warnings is frankly iffy
 
-It works by fiddling with the (global) %warnings::Bits variable, and that's
-fine for run-time warnings, but compile-time warnings like e.g.
-C<experimental::smartmatch> will still happen, even if you think
-you've disabled them.
+The best I can say about its method of messing with the source code of
+imported modules is that at least its modifications shouldn't stack with
+other source filters, so the degree of weirdness and potential insanity
+should be reduced to a manageable level.
 
 =back
 
@@ -316,7 +386,8 @@ sub _set_bit_mask {
 sub _is_bit_set {
     my ($mask, $bit_num) = @_;
 
-    return vec($mask, int($bit_num / 8), 8) & (1 << ($bit_num % 8));
+    my $smallest_bit_num = $bit_num % 8;
+    return vec($mask, int($bit_num / 8), 8) & (1 << $smallest_bit_num);
 }
 
 sub _dump_mask {
@@ -341,6 +412,106 @@ sub _check_warning_category {
     return 1;
 }
 
+sub _thwart_module {
+    my ($package, %args) = @_;
+
+    # There are two ways of thwarting modules: the usual Perlish way,
+    # and one special mode for Moose, which generates its own import
+    # sub which can't be wrapped.
+    my $module = $args{thwart_module};
+    my $mode = 'append_sub';
+    if ($module eq 'Moose') {
+        $module = 'Moose::Exporter';
+        $mode   = 'append_statement';
+    }
+    my $filename = $module;
+    $filename =~ s{::}{/}g;
+    $filename .= '.pm';
+    unshift @INC, sub {
+        my ($this_coderef, $use_filename) = @_;
+        return if $use_filename ne $filename;
+
+        # Find the source of the module we're looking for.
+        # This will fail if the module is itself being loaded by a
+        # coderef in @INC, say, but should work for the vast, vast
+        # majority of cases.
+        my $source_fh = $package->_find_module_source($use_filename)
+            or do {
+            croak "You asked me to thwart $args{thwart_module}"
+                . " but I can't find $use_filename anywhere in @INC";
+            };
+        my $source;
+        {
+            local $/;
+            $source = <$source_fh>;
+        }
+
+        # Work out what we're going to inject into this source code.
+        my @warnings
+            = ref($args{warning} eq 'ARRAY')
+            ? @{ $args{warning} }
+            : $args{warning};
+        my $source_code_unimport;
+        for my $warning (@warnings) {
+            $source_code_unimport .= qq{warnings->unimport("$warning");\n};
+        }
+        my $injection_warning_start = "### Code injected by $package";
+        my $injection_warning_end   = "### End of code injected by $package";
+
+        # We might be adding an extra import sub.
+        if ($mode eq 'append_sub') {
+            my $extra_code = <<EXTRACODE;
+$injection_warning_start
+my \$__warnings_everywhere_orig_import = \\\&${module}::import;
+{
+    no warnings 'redefine';
+    *${module}::import = sub {
+        \$__warnings_everywhere_orig_import->(\@_);
+        $source_code_unimport
+    }
+}
+$injection_warning_end
+1;
+EXTRACODE
+
+            # This wants to go either at the end, or before __END__ or
+            # __DATA__
+            $source =~ s/^ (__ (?: END|DATA ) __) $/$extra_code$1/xsm
+                or $source .= $extra_code;
+
+        # Or we might be adding our import statements immediately
+        # after the call to import.
+        } elsif ($mode eq 'append_statement') {
+            $source =~ s{
+                ( warnings->import; \n )
+            }{
+$1
+$injection_warning_start
+$source_code_unimport
+$injection_warning_end
+            }x;
+        }
+
+        # Right, return this modified source code.
+        open (my $fh_source, '<', \$source);
+        return $fh_source;
+    };
+}
+
+sub _find_module_source {
+    my ($package, $use_filename) = @_;
+
+    for my $dir (grep { !ref($_) } @INC) {
+        my $full_path = File::Spec->catfile($dir, $use_filename);
+        if (-e $full_path) {
+            open (my $fh, '<', $full_path);
+            return $fh;
+        }
+    }
+    return;
+}
+
+
 =back
 
 =head1 TO DO
